# pylint: disable=W0232, C1001, C0111, R0913, E1101, W0212
# TODO end_contract; record all payments
from __future__ import print_function
from builtins import str
from builtins import object
from abcEconomics.notenoughgoods import NotEnoughGoods
from random import shuffle
from abcEconomics.trade import get_epsilon
from .contracts import Contracts


epsilon = get_epsilon()


class Contract(object):
    __slots__ = ['sender_group', 'sender_id', 'deliver_good_group',
                 'deliver_good_id', 'pay_group', 'pay_id', 'good', 'quantity',
                 'price', 'end_date', 'delivered', 'paid', 'id', 'round']

    def __init__(self, sender_group, sender_id, deliver_good_group,
                 deliver_good_id, pay_group, pay_id, good, quantity, price,
                 end_date, id, round):
        self.sender_group = sender_group
        self.sender_id = sender_id
        self.deliver_good_group = deliver_good_group
        self.deliver_good_id = deliver_good_id
        self.pay_group = pay_group
        self.pay_id = pay_id
        self.good = good
        self.quantity = quantity
        self.price = price
        self.end_date = end_date
        self.delivered = []
        self.paid = []
        self.id = id
        self.round = round

    def __str__(self):
        return str(('sender', self.sender_group, self.sender_id, 'deliver', self.deliver_good_group,
                    self.deliver_good_id, self.pay_group, self.pay_id, self.good, self.quantity, self.price,
                    self.end_date, self.id, self.delivered, self.paid))


class Contracting(object):
    """ This is a class, that allows you to create contracts. For example a
    work contract. One agent commits to deliver a good or service for a set
    amount of time.

    For example you have a firm and a worker class. 'Labor' is set as a service
    meaning that it lasts not longer than one round and the worker how has an
    adult gets one unit of labor every round see: :meth:`abcEconomics.declare_service`.
    The firm offers a work contract, the worker responds. Every round the
    worker delivers the labor and the firm pays.::

        class Firm(abcEconomics.Agent, abcEconomics.Contract)
            def request_offer(self):
                if self.round % 10 == 0:
                    self.given_contract = self.request_contract('contractbuyer', 0,
                                                                good='labor',
                                                                quantity=5,
                                                                price=10,
                                                                duration=10 - 1)

            def deliver_or_pay(self):
                self.pay_contract('labor')

        class Worker(abcEconomics.Agent, abcEconomics.Contract):
            def init(self):
                self.create('adult', 1)

            def accept_offer(self):
                contracts = self.get_contract_requests('labor')
                for contract in contracts:
                    if contract.price < 5:
                        self.accepted_contract = self.accept_contract(contract)

            def deliver_or_pay(self):
                self.deliver('labor')

    Firms and workers can check, whether they have been paid/provided with
    labor using the :meth:`is_paid` and :meth:`is_delivered` methods.

    The worker can also initiate the transaction by requesting a contract with
    :meth:`make_contract_offer`.

    A contract has the following fields:

                 sender_group:

                 sender_id:

                 deliver_group:

                 deliver_id:

                 pay_group:

                 pay_id:

                 good:

                 quantity:

                 price:

                 end_date:

                 makerequest:
                    'm' for make_contract_offer and 'r' for request_contract

                 id:
                    unique number of contract

    """
    def _add_contracts_list(self):
        self.contracts = Contracts(self.name)

    def offer_good_contract(self, receiver_group, receiver_id, good, quantity, price, duration):
        """This method offers a contract to provide a good or service to the
        receiver. For a given time at a given price.

        Args:

            receiver_group:
                group to receive the good
            receiver_id:
                group to receive the good
            good:
                the good or service that should be provided
            quantity:
                the quantity that should be provided
            price:
                the price of the good or service
            duration:
                the length of the contract, if duration is None or not set,
                the contract has no end date.

        Example::

            self.given_contract = self.make_contract_offer('firm', 1, 'labor', quantity=8, price=10, duration=10 - 1)
        """
        quantity = bound_zero(quantity)
        if duration is None:
            end_date = None
        else:
            end_date = duration + self.round

        offer = Contract(sender_group=self.group,
                         sender_id=self.id,
                         deliver_good_group=self.group,
                         deliver_good_id=self.id,
                         pay_group=receiver_group,
                         pay_id=receiver_id,
                         good=good,
                         quantity=quantity,
                         price=price,
                         end_date=end_date,
                         id=self._offer_counter(),
                         round=self.round)
        self._send(receiver_group, receiver_id, '!o', offer)
        self._contract_offers_made[offer.id] = offer
        return offer

    def request_good_contract(self, receiver_group, receiver_id, good, quantity, price, duration):
        """This method requests a contract to provide a good or service to the
        sender. For a given time at a given price. For example a job
        advertisement.

        Args:

            receiver_group:
                group of the receiver
            receiver_id:
                id of the receiver
            good:
                the good or service that should be provided
            quantity:
                the quantity that should be provided
            price:
                the price of the good or service
            duration:
                the length of the contract, if duration is None or not set,
                the contract has no end date.
        """
        quantity = bound_zero(quantity)
        if duration is None:
            end_date = None
        else:
            end_date = duration + self.round

        offer = Contract(sender_group=self.group,
                         sender_id=self.id,
                         pay_group=self.group,
                         pay_id=self.id,
                         deliver_good_group=receiver_group,
                         deliver_good_id=receiver_id,
                         good=good,
                         quantity=quantity,
                         price=price,
                         end_date=end_date,
                         id=self._offer_counter(),
                         round=self.round)
        self._send(receiver_group, receiver_id, '!o', offer)
        self._contract_offers_made[offer.id] = offer
        return offer

    def get_contract_offers(self, good, descending=False):
        """ Returns all contract offers and removes them. The contract
        are ordered by price (ascending), when tied they are randomized.

        Args:
            good:
                good that underlies the contract

            descending(bool,default=False):
                False for descending True for ascending by price

        Returns:
            list of contract offers ordered by price
        """
        ret = self._contract_offers[good]
        del self._contract_offers[good]
        shuffle(ret)
        ret.sort(key=lambda objects: objects.price, reverse=descending)
        return ret

    def accept_contract(self, contract, quantity=None):
        """ Accepts the contract. The contract is completely accepted, when
        the quantity is not given. Or partially when quantity is set.

        Args:

            contract:
                the contract in question, received with get_contract_requests or
                get_contract_offers

            quantity (optional):
                the quantity that is accepted. Defaults to all.
        """
        if quantity is None:
            quantity = contract.quantity
        else:
            contract.quantity = min(contract.quantity, quantity)
            assert quantity < contract.quantity + \
                epsilon * max(quantity, contract.quantity)
            if quantity > contract.quantity:
                quantity = contract.quantity

        if contract.pay_group == self.group and contract.pay_id == self.id:
            self.contracts._contracts_pay[contract.good][contract.id] = contract
            self._send(contract.sender_group,
                       contract.sender_id, '_ac', contract)
        else:
            self.contracts._contracts_deliver[contract.good][contract.id] = contract
            self._send(contract.sender_group,
                       contract.sender_id, '_ac', contract)
        return contract

    def deliver_contract(self, contract):
        """ delivers on a contract """
        assert contract.deliver_good_group == self.group and contract.deliver_good_id == self.id
        quantity = contract.quantity
        available = self._haves[contract.good]
        if quantity > available + epsilon + epsilon * max(quantity, available):
            raise NotEnoughGoods(self.name, contract.good,
                                 quantity - available)
        if quantity > available:
            quantity = available

        self._haves[contract.good] -= quantity
        self._send(contract.pay_group, contract.pay_id, '_dp', contract)

    def pay_contract(self, contract):
        """ delivers on a contract """
        assert contract.pay_group == self.group and contract.pay_id == self.id
        money = contract.quantity * contract.price
        available = self._haves['money']
        if money > available + epsilon + epsilon * max(money, available):
            raise NotEnoughGoods(self.name, 'money', money - available)
        if money > available:
            money = available

        self._haves['money'] -= money
        self._send(contract.deliver_good_group,
                   contract.deliver_good_id, '_dp', contract)

    def contracts_to_deliver(self, good):
        return list(self.contracts._contracts_deliver[good].values())

    def contracts_to_receive(self, good):
        return list(self.contracts._contracts_pay[good].values())

    def contracts_to_deliver_all(self):
        ret = {}
        for good in self.contracts._contracts_deliver:
            ret[good] = list(self.contracts._contracts_deliver[good].values())
        return ret

    def contracts_to_receive_all(self):
        ret = {}
        for good in self.contracts._contracts_pay:
            ret[good] = list(self.contracts._contracts_pay[good].values())
        return ret

    def end_contract(self, contract):
        if contract.id in self.contracts._contracts_deliver[contract.good]:
            self._send(contract.pay_group, contract.pay_id,
                       '!d', ('r', contract.good, contract.id))
            del self.contracts._contracts_deliver[contract.good][contract.id]
        elif contract.id in self.contracts._contracts_pay[contract.good]:
            self._send(contract.deliver_good_group, contract.deliver_good_id,
                       '!d', ('d', contract.good, contract.id))
            del self.contracts._contracts_pay[contract.good][contract.id]
        else:
            raise Exception("Contract not found")

    def was_paid_this_round(self, contract):
        return contract.paid[-1] == self.round

    def was_delivered_this_round(self, contract):
        return contract.delivered[-1] == self.round

    def was_paid_last_round(self, contract):
        return self.round - 1 in contract.paid

    def was_delivered_last_round(self, contract):
        return self.round - 1 in contract.delivered

    def calculate_netvalue(self, prices={},
                           parameters={},
                           value_functions={}):
        return (self._haves.calculate_netvalue(prices) +
                self.contracts.calculate_netvalue(parameters, value_functions))

    def calculate_assetvalue(self, prices={},
                             parameters={},
                             value_functions={}):
        return (self._haves.calculate_assetvalue(prices) +
                self.contracts.calculate_assetvalue(parameters,
                                                    value_functions))

    def calculate_liablityvalue(self, prices={},
                                parameters={},
                                value_functions={}):
        return (self._haves.calculate_liablityvalue(prices) +
                self.contracts.calculate_liablityvalue(parameters,
                                                       value_functions))

    def calculate_valued_assets(self, prices={},
                                parameters={},
                                value_functions={}):
        return (self._haves.calculate_valued_assets(prices) +
                self.contracts.calculate_valued_assets(parameters,
                                                       value_functions))

    def calculate_valued_liablities(self, prices={},
                                    parameters={},
                                    value_functions={}):
        return (self._haves.calculate_valued_liablities(prices) +
                self.contracts.calculate_valued_liablities(parameters,
                                                           value_functions))


def bound_zero(x):
    """ asserts that variable is above zero, where foating point imprecission is accounted for,
    and than makes sure it is above 0, without floating point imprecission """
    assert x > - epsilon, \
        '%.30f is smaller than 0 - epsilon (%.30f)' % (x, - epsilon)
    if x < 0:
        return 0
    else:
        return x




import os
try:
    from setuptools import setup
    from setuptools import Extension
except ImportError:
    from distutils.core import setup
    from distutils.extension import Extension
try:
    from Cython.Distutils import build_ext
except ImportError:
    from distutils.command.build_ext import build_ext
from distutils.errors import CCompilerError, DistutilsExecError, \
    DistutilsPlatformError
import platform


class TXEntension(build_ext):
    # This class allows C extension building to fail.
    def run(self):
        try:
            build_ext.run(self)
        except DistutilsPlatformError:
            raise Exception("BuildFailed")

    def build_extension(self, ext):
        try:
            build_ext.build_extension(self, ext)
        except (CCompilerError, DistutilsExecError, DistutilsPlatformError):
            pass  # raise BuildFailed()


cmdclass = {}
ext_modules = []

install_requires = ['flexx >= 0.4.1',
                    'future',
                    'dataset == 0.8']


readthedocs = os.environ.get('READTHEDOCS') == 'True'

if not readthedocs:
    try:
        ext_modules += [
            Extension("abcEconomics.trade", ["abcEconomics/trade.pyx"]),
            Extension("abcEconomics.logger.online_variance", ["abcEconomics/logger/online_variance.pyx"]),
        ]
        cmdclass.update({'build_ext': TXEntension})
    except ImportError:
        ext_modules += [
            Extension("abcEconomics.trade", ["abcEconomics/trade.c"]),
            Extension("abcEconomics.logger.online_variance", ["abcEconomics/logger/online_variance.c"]),
        ]

    if not platform.python_implementation() == "PyPy":
        install_requires += ['numpy >= 1.10.2p']
        if ('APPVEYOR' not in os.environ) or ('TRAVIS' not in os.environ):
            install_requires += ['pandas >= 0.17.1',
                                 'bokeh == 0.12.16',
                                 'tornado == 4.3']


version = '0.9.6b0'


setup(name='abcEconomics',
      version=version,
      author='Davoud Taghawi-Nejad',
      author_email='Davoud@Taghawi-Nejad.de',
      description='Agent-Based Complete Economy modelling platform',
      url='https://github.com/AB-CE/abce.git',
      package_dir={'abcEconomics': 'abcEconomics',
                   'abcEconomics.gui': 'abcEconomics/gui',
                   'abcEconomics.agents': 'abcEconomics/agents',
                   'abcEconomics.contracts': 'abcEconomics/contracts',
                   'abcEconomics.logger': 'abcEconomics/logger',
                   },
      packages=['abcEconomics'],
      long_description=open('README.rst').read(),
      install_requires=install_requires,
      include_package_data=True,
      ext_modules=ext_modules,
      cmdclass=cmdclass)




# Copyright 2012 Davoud Taghawi-Nejad
#
# Module Author: Davoud Taghawi-Nejad
# abcEconomics is open-source software. If you are using abcEconomics for your research you are
# requested the quote the use of this software.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License and quotation of the
# author. You may obtain a copy of the License at
#       http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations under
# the License.
# pylint: disable=W0201
""" The Firm class allows you to set up firm agents with
complex or several production functions. While the simple Firm automatically
handles one technology, Firm allows you to manage several
technologies manually.

The create_* functions allow you to create a technology and assign it to
a variable. :meth:`abcEconomics.Firm.produce` and similar
methods use this variable to produce with the according technology.
"""
import operator
from functools import reduce
from collections import ChainMap
from abcEconomics import NotEnoughGoods
from ..trade import get_epsilon
epsilon = get_epsilon()


class Firm:
    """ With :code:`self.produce` a firm produces a good using production functions.
    For example the following farm has a cobb-douglas production function:

    class Farm(abcEconomics.Agent, abcEconomics.Firm):
        def init(self):
            self.production_function = create_cobb_douglas({'land': 0.7,
                                                            'capital': 0.1,
                                                            'labor': 0.2})

        def firming(self):
            self.produce(self.production_function, {{'land': self['land'],
                                                     'capital': self['capital'],
                                                     'labor': 2}})

    Production functions can be auto generated with:
        - py:meth:`~abcEconomicsagent.Firm.create_cobb_douglas` or
        - py:meth:`~abcEconomicsagent.Firm.create_ces` or
        - py:meth:`~abcEconomicsagent.Firm.create_leontief`

    or specified by hand:

    A production function looks like this::

        def production_function(wheels, steel, stearing_wheels, machines):
            result = {'car': min(wheels / 4, steel / 10, stearing_wheels),
                      'wheels': 0,
                      'steel': 0,
                      'steering_wheels': 0,
                      'machine': machine * 0.9}
            return result

    Or more readably like this:

        def production_function(wheels, steel, stearing_wheels, machines):
            car = min(wheels / 4, steel / 10, stearing_wheels)
            wheels = 0
            steel = 0
            stearing_wheels = 0
            machine = machine * 0.9
            return locals()

    This production function, produces one car for every four wheels, 10 tonnes of steel
    and one stearing_wheel, it also requires one machine.  Wheels, steel and stearing_wheels
    are completely used. The plant is not used and the machine depreciates by 10%.production.

    A production function can also produce multiple goods. The last line :code:`return locals()`,
    can not be omitted. It returns all variables you define in this function as a dictionary.
    """

    def produce(self, production_function, input_goods, results=False):
        """ Produces output goods given the specified amount of inputs.

        Transforms the Agent's goods specified in input goods
        according to a given production_function to output goods.
        Automatically changes the agent's belonging. Raises an
        exception, when the agent does not have sufficient resources.

        Args:
            production_function:
                A production_function produced with
                py:meth:`~abcEconomicsagent.Firm.create_production_function`,
                py:meth:`~abcEconomicsagent.Firm.create_cobb_douglas` or
                py:meth:`~abcEconomicsagent.Firm.create_leontief`

            input goods dictionary or list:
                dictionary containing the amount of input good used for the production or
                a list of all goods that get completely used.

            results:
                If True returns a dictionary with the used and produced goods.

        Raises:
            NotEnoughGoods:
                This is raised when the goods are insufficient.

        Example::

            car = {'tire': 4, 'metal': 2000, 'plastic':  40}
            bike = {'tire': 2, 'metal': 400, 'plastic':  20}
            try:
                self.produce(car_production_function, car)
            except NotEnoughGoods:
                A.produce(bike_production_function, bike)

            self.produce(car_production_function, ['tire', 'metal', 'plastic'])  # produces using all goods

        """
        if not isinstance(input_goods, dict):
            input_goods = {good: self[good] for good in input_goods}

        result = production_function(**input_goods)

        for good, quantity in input_goods.items():
            if self._inventory.haves[good] - quantity + result.get(good, 0) < -epsilon:
                raise NotEnoughGoods

        for good, quantity in input_goods.items():
            self._inventory.haves[good] -= quantity

        for good, quantity in result.items():
            self._inventory.haves[good] += quantity

        if results:
            return {good: result.get(good, 0) - input_goods.get(good, 0)
                    for good in ChainMap(input_goods, result).keys()}

    def create_cobb_douglas(self, output, multiplier, exponents):
        """ creates a Cobb-Douglas production function

        A production function is a production process that produces the
        given input  goods according to the Cobb-Douglas formula to the output
        good.
        Production_functions are than used as an argument in produce,
        predict_vector_produce and predict_output_produce.

        Args:

            'output':
                Name of the output good

            multiplier:
                Cobb-Douglas multiplier

            {'input1': exponent1, 'input2': exponent2 ...}:
                dictionary containing good names 'input' and corresponding exponents

        Returns:

            A production_function that can be used in produce etc.

        Example:

            def init(self):
                self.plastic_production_function = create_cobb_douglas('plastic', {'oil' : 10, 'labor' : 1}, 0.000001)

            ...

            def producing(self):
                self.produce(self.plastic_production_function, {'oil' : 20, 'labor' : 1})

        """
        def production_function(**goods):
            ret = multiplier * reduce(operator.mul,
                                      [goods[name] ** exponent
                                       for name, exponent in exponents.items()])
            return {output: ret}
        return production_function

    def create_ces(self, output, gamma, multiplier=1, shares=None):
        """ creates a CES production function

        A production function is a production process that produces the
        given input  goods according to the CES formula to the output
        good:


        :math:`Q = F \\cdot \\left[\\sum_{i=1}^n a_{i}X_{i}^{\\gamma}\\ \\right]^{\\frac{1}{\\gamma}}`


        Production_functions are than used as an argument in produce,
        predict_vector_produce and predict_output_produce.

        Args:

            'output':
                Name of the output good

            gamma:
                elasticity of substitution :math:`= s =\\frac{1}{1-\\gamma}`

            multiplier:
                CES multiplier :math:`F`

            shares:
                :math:`a_{i}` = Share parameter of input i, :math:`\\sum_{i=1}^n a_{i} = 1`
                when share_parameters is not specified all inputs are weighted equally and
                the number of inputs is flexible.

        Returns:

            A production_function that can be used in produce etc.

        Example::

            self.stuff_production_function = create_ces('stuff', gamma=0.5, multiplier=1,
                                                        shares={'labor': 0.25, 'stone':0.25, 'wood':0.5})
            self.produce(self.stuff_production_function, {'stone' : 20, 'labor' : 1, 'wood': 12})

        """
        if shares is None:
            def production_function(**goods):
                a = 1 / len(goods)
                ret = multiplier * sum([a * goods[name] ** gamma
                                        for name in goods]) ** (1 / gamma)
                return {output: ret}
        else:
            def production_function(**goods):
                ret = multiplier * sum([share * goods[name] ** gamma
                                        for name, share in shares.items()]) ** (1 / gamma)
                return {output: ret}

        return production_function

    def create_leontief(self, output, utilization_quantities):
        """ creates a Leontief production function

        A production function is a production process that produces the
        given input  goods according to the Leontief formula to the output
        good.
        Production_functions are than used as an argument in produce,
        predict_vector_produce and predict_output_produce.

        Args:

            'output':
                Name of the output good

            multiplier:
                dictionary of multipliers it min(good1 * a, good2 * b, good3 * c...)

            {'input1': exponent1, 'input2': exponent2 ...}:
                dictionary containing good names 'input' and corresponding exponents

        Returns:

            A production_function that can be used in produce etc.

        Example:
        self.car_production_function = create_leontief('car', {'wheel' : 4, 'chassi' : 1})
        self.produce(self.car_production_function, {'wheel' : 20, 'chassi' : 5})

        """
        def production_function(**goods):
            ret = min([goods[name] * factor for name, factor in utilization_quantities.items()])
            return {output: ret}

        return production_function



import abcEconomics
import dataset


class CustomLogging:
    def __init__(self, dbprotocolname, tablename):
        self.db = dataset.connect(dbprotocolname)
        self.table = self.db[tablename]

    def write_everything(self, **kveverything):
        self.table.insert(kveverything)

    def close(self):
        self.db.commit()


class MyAgent(abcEconomics.Agent):
    def write(self):
        self.custom_log('write_everything', name='joe', m=5)


def main(processes, rounds):
    sim = abcEconomics.Simulation(name='mysim', processes=processes, dbplugin=CustomLogging, dbpluginargs=['sqlite:///:memory:', 'sometable'])

    myagents = sim.build_agents(MyAgent, 'myagent', number=5)

    for i in range(rounds):
        sim.advance_round(i)
        myagents.write()

    sim.finalize()


if __name__ == '__main__':
    main(1, 5)
    main(4, 5)



# Copyright 2012 Davoud Taghawi-Nejad
#
# Module Author: Davoud Taghawi-Nejad
#
# abcEconomics is open-source software. If you are using abcEconomics for your research you
# are requested the quote the use of this software.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License and quotation of the
# author. You may obtain a copy of the License at
#       http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations under
# the License.
"""
The :class:`abcEconomicsagent.Agent` class is the basic class for creating your agent.
It automatically handles the
possession of goods of an agent. In order to produce/transforme goods you
need to also subclass the :class:`abcEconomicsagent.Firm` [1]_ or to create a
consumer the :class:`abcEconomicsagent.Household`.

For detailed documentation on:

Trading:
    see :class:`abcEconomicsagent.Trade`

Logging and data creation:
    see :class:`abcEconomicsagent.Database` and :doc:`simulation_results`

Messaging between agents:
    see :class:`abcEconomicsagent.Messenger`.

.. autoexception:: abcEconomics.NotEnoughGoods

.. [1] or :class:`abcEconomicsagent.FirmMultiTechnologies` for  complex technologies.
"""
from collections import OrderedDict
import re

import abcEconomics


class Logger:
    """ The Logger class """
    def __init__(self, id, agent_parameters, simulation_parameters):
        super(Logger, self).__init__(id, agent_parameters, simulation_parameters)
        # unpack simulation_parameters
        database = simulation_parameters['database']
        trade_logging = simulation_parameters['trade_logging']

        self.database_connection = database

        self._data_to_log_1 = {}
        self._data_to_observe = {}

        if hasattr(abcEconomics, 'conditional_logging'):
            self.conditional_logging = True
            self.log_rounds = abcEconomics.conditional_logging
        else:
            self.conditional_logging = False
        self.log_this_round = True

        self.trade_logging = {'individual': 1,
                              'group': 2,
                              'off': 0}[trade_logging]

    def log(self, action_name, data_to_log):
        """ With log you can write the models data. Log can save variable
        states and and the working of individual functions such as production,
        consumption, give, but not trade(as its handled automatically). Sending
        a dictionary instead of several using several log statements with a
        single variable is faster.

        Args:
            'name'(string):
                the name of the current action/method the agent executes

            data_to_log:
                a variable or a dictionary with data to log in the the database

        Example::

            self.log('profit', profit)

            self.log('employment_and_rent',
                     {'employment': self['LAB'],
                     'rent': self['CAP'],
                     'composite': self.composite})

            self.log(self.produce_use_everything())

        """
        if self.log_this_round:
            self.database_connection.put(
                ("log",
                 self._str_round,
                 self.group,
                 self._str_name,
                 re.sub('[^0-9a-zA-Z_]', '', action_name),
                 data_to_log))

    def _common_log(self, variables, possessions, functions, lengths):
        ret = OrderedDict()
        for var in variables:
            ret[var] = self.__dict__[var]
        for pos in possessions:
            ret[pos] = self._inventory[pos]
        for name, func in functions.items():
            ret[name] = func(self)
        for length in lengths:
            ret['len_' + length] = len(self.__dict__[length])
        return ret

    def _agg_log(self, variables, possessions, functions, lengths):
        if self.log_this_round:
            data_to_write = self._common_log(variables,
                                             possessions,
                                             functions,
                                             lengths)
            self.database_connection.put(["snapshot_agg",
                                          self._str_round,
                                          self.group,
                                          data_to_write])

    def _panel_log(self, variables, possessions, functions, lengths, serial):
        if self.log_this_round:
            data_to_write = self._common_log(variables,
                                             possessions,
                                             functions,
                                             lengths)
            self.database_connection.put(("panel_log",
                                          self._str_round,
                                          self.group,
                                          self._str_name,
                                          serial,
                                          data_to_write))

    def custom_log(self, method, *args, **kwargs):
        """ send custom logging commands to database plugin, see :ref:`Database Plugins` """
        self.database_connection.put([method, args, kwargs])



from __future__ import division
from __future__ import print_function
import abcEconomics
from abcEconomics.agents import Firm


class ExpiringCapital(abcEconomics.Agent, Firm):
    def init(self, rounds, xcapital):
        self.last_round = rounds - 1
        self.create('xcapital', 1)
        print('============', xcapital)

    def one(self):
        pass

    def two(self):
        pass

    def three(self):
        pass

    def go(self):

        if self.round == 2:
            self.create('xcapital', 10)

        if self.round == 0:
            assert self['xcapital'] == 1
        if self.round == 1:
            assert self['xcapital'] == 1
        if self.round == 2:
            assert self['xcapital'] == 11
        if self.round == 3:
            assert self['xcapital'] == 11
        if self.round == 4:
            assert self['xcapital'] == 11
        if self.round == 5:
            assert self['xcapital'] == 10

    def clean_up(self):
        pass

    def all_tests_completed(self):
        if self.round == self.last_round and self.id == 0:
            print("simple ExpiringCapital \tOK")



import abcEconomics
from flexx import ui


class LoadForm(ui.Widget):
    def init(self):
        with ui.Widget(style="overflow-y: scroll"):
            with ui.VBox(style="overflow-y: scroll") as self.vbox:
                name_descriptions = [(d['name'], d['description'])
                                     for d in abcEconomics.parameter_database.all()]
                for name, desc in name_descriptions:
                    with ui.GroupWidget():
                        btn = ui.Button(title=name, text=name)
                        delete = ui.Button(title=name, text='(del)')
                        ui.Label(text=desc)

                    btn.connect('mouse_click', self.wdg)
                    delete.connect('mouse_click', self.delete)

    def update(self, event):
        with self.vbox:
            with ui.GroupWidget():
                btn = ui.Button(title=event['name'],
                                text=event['name'])
                delete = ui.Button(title=event['name'], text='del')
                ui.Label(text=event['description'])
            btn.connect('mouse_click', self.wdg)
            delete.connect('mouse_click', self.delete)

    def wdg(self, event):
        self.emit('load', {'name': event['source'].title})

    def delete(self, event):
        abcEconomics.parameter_database.delete(name=event['source'].title)
        event['source'].text = 'deleted'




import abcEconomics
from abcEconomics.agents import Firm


class GiveExpiringCapital(abcEconomics.Agent, Firm):
    def init(self, rounds):
        self.last_round = rounds - 1
        if self.id == 0:
            self.create('xcapital', 10)
            assert self['xcapital'] == 10, self['xcapital']

    def one(self):
        if self.id == 0:
            assert self['xcapital'] == 10, self['xcapital']
            self.give('giveexpiringcapital', 1, 'xcapital', 10)
            assert self['xcapital'] == 0, self['xcapital']

    def two(self):
        if self.id == 1:
            assert self['xcapital'] == 10, self['xcapital']
            self.give('giveexpiringcapital', 0, 'xcapital', 10)
            assert self['xcapital'] == 0, self['xcapital']

    def three(self):
        pass

    def clean_up(self):
        pass

    def all_tests_completed(self):
        if self.round == self.last_round and self.id == 0:
            print("Give ExpiringCapital \tOK")
